#!/usr/local/bin/perl
############
# killtnef #
############

#History:
#          1.0.0.3,  fixed killtnef so out file truly is used if specified.  Before only msgs with mstnef content
#                    was sent to the out file, the rest went to STDOUT  :-(
#

use strict;
use Convert::TNEF;
use MIME::Parser;
use Getopt::Long;
use Pod::Usage;

my $out_file='STDOUT';
my $in_file;
my $help=0;
my $man=0;
my $killtnef_version=0;

my $VERSION = 1.0;

GetOptions ('in=s' => \$in_file,      'out=s' => \$out_file,
            'help|?' => \$help,                  man => \$man, 
            'version' =>\$killtnef_version     ) or pod2usage(2);

 pod2usage(1) if $help;
 pod2usage(-exitstatus => 0, -verbose => 2) if $man;
 
 if ($killtnef_version) {
  print 'Killtnef - version 1.0.0.3, 99.999% the work of Andrew Orr and Lester Hightower';
  print ' .001% by Dennis Heidner';
  exit;}

if (!defined $in_file) {die '--in missing on command line';}
open (INFILE, "< $in_file") or die 'Infile does not exist';

if (!defined $out_file) {die '--out missing on command line';}
open (OUTFILE, "> $out_file")  or die 'Unable to open output file';

my $message;
while (defined(my $line = <INFILE>)) {
  my $msg_bound;
  if ($line =~ /^(From\s+.+)\r?\n/ and length($message) or eof) {
    parse_message(\$message, $msg_bound); # All the action happens here...
    $message = "";
    $msg_bound=$line;
  } elsif ($line =~ /^(From\s+.+)\r?\n/) {
    $msg_bound=$line;  # The first 
  } 
  $message.=$line;
}
close (INFILE);
if (defined $out_file) {close (OUTFILE)};
exit;


# Subroutines ###############################################################

sub parse_message {
  my $msg_body = shift @_;
  my $msg_bound = shift @_;
  my $mime_parser = new MIME::Parser;

  # This module likes to use tmp files, but I try to stop it here.
  $mime_parser->use_inner_files(1);
  $mime_parser->output_to_core(1);
  my $ent = $mime_parser->parse_data($$msg_body);
  my $num_parts=$ent->parts;

  # Determine if we have a MIME w/ms-tnef and act accordingly.
  if ( ($num_parts < 1) || ($$msg_body !~ /ms-tnef/i) ) {
    print OUTFILE "$$msg_body";  
  } else {
    # Get the head info
    my $head = $ent->head;
    my $ReturnPath = $head->get('Return-Path');
    my @all_received = $head->get('Received');
    my $Date = $head->get('Date');
    my $From = $head->get('From');
    my $XSender = $head->get('X-Sender');
    my $To = $head->get('To');
    my $Subject = $head->get('Subject');
    my $MessageID = $head->get('Message-ID');

    # Build a new MIME message based on the one we are examining
    # - LHH: it would probably be better to build this $new_ent
    #        using $ent->head as the basis, thus getting *all* of
    #        the headers, instead of just these few.  We only needed
    #        these few headers for the project this script was
    #        originally written for, but if someone wants to change
    #        this and submit a patch, that would be great.
    my $new_ent = MIME::Entity->build(
				'Type'        => "multipart/mixed",
				'Return-Path' => $ReturnPath,
				'Date'        => $Date,
				'From'        => $From,
				'X-Sender'    => $XSender,
				'To'          => $To,
				'Subject'     => $Subject,
				'Message-ID'  => $MessageID
                                  );
    my $new_head=$new_ent->head;
    foreach my $received (@all_received) {
      $new_head->add('Received', $received);
    }

    # Loop over each MIME part adding each to the new message
    foreach my $mime_part_i (0 .. ($num_parts - 1)) {
      my $ent_part=$ent->parts($mime_part_i);
      if ($ent_part->mime_type =~ /ms-tnef/i )  {
        add_tnef_parts($ent_part, $new_ent);
      } else {
        $new_ent->add_part($ent_part);
      } 
    }

    # Set the preamble and epilogue equal to the original
    $new_ent->preamble($ent->preamble);
    $new_ent->epilogue($ent->epilogue);

    # Print the newly constructed MIME message
    print OUTFILE "$msg_bound";    
    print OUTFILE $new_ent->stringify;
  }
}

sub add_tnef_parts {
  my $ent = shift;
  my $new = shift;

  # LHH: This should probably be loaded from /etc/mime.types, or something.
  #      A command line argument to this script pointing it at a mime.types
  #      file is probably appropriate.
  # DLH: I've added about thirty additional MIME types
  my %content_type = (
                       aif  => "audio/x-aiff",
                       aiff => "audio/x-aiff",
                       aifc => "audio/x-aiff",
                       arj  => "application/x-arj-compressed",
                       asc  => "text/plain",
                       avi  => "video/x-msvideo",
                       txt  => "text/plain",
                       htm  => "text/html",
                       html => "text/html",
                       doc  => "application/msword",
                       dll  => "application/octet-stream",
                       bat  => "application/x-msdos-program",
                       com  => "application/x-msdos-program",
                       csh  => "application/x-csh",
                       csv  => "text/comma-separated-values",
                       eps  => "application/postscript",
                       exe  => "application/x-msdos-program",
                       gz   => "application/x-gunzip",
                       mp2  => "video/mpeg",
                       mpe  => "video/mpeg",
                       mpeg => "video/mpeg",
                       mpg  => "video/mpeg",
                       man  => "application/x-troff-man",
                       mid  => "audio/midi",
                       midi => "audio/midi",
                       mov  => "video/quicktime",
                       xls  => "application/vnd.ms-excel",
                       ppt  => "application/vnd.ms-powerpoint",
                       pdf  => "application/pdf",
                       rtf  => "application/rtf",
                       zip  => "application/zip",
                       obj  => "application/octet-stream",
                       pgp  => "application/pgp",
                       pl   => "application/x-perl",
                       pm   => "application/x-perl",
                       ps   => "application/ps",
                       qt   => "video/quicktime",
                       rar  => "application/x-rar-compressed",
                       ras  => "image/x-cmu-raster",
                       rpm  => "application/x-redhat-package-manager",
                       t    => "application/x-troff",
                       tr   => "application/x-troff",
                       tar  => "application/x-tar",
                       tcl  => "application/x-tcl",
                       tgz  => "application/x-tar-gz",
                       tsv  => "text/tab-separated-values",
                       vrm  => "x-world/x-vrml",
                       vrml => "x-world/x-vrml",
                       vrl  => "x-world/x-vrml",
                       wav  => "audio/x-wav",
                       wpd  => "application/wordperfect",
                       wks  => "application/vnd.ms-works",
                       xml  => "text/xml",
                       bin  => "application/octet-stream",
                       bmp  => "image/bmp",
                       jpe  => "image/jpeg",
                       jpeg => "image/jpeg",
                       jpg  => "image/jpeg",
                       raw  => "image/raw",
                       gif  => "image/gif",
                       tif  => "image/tif",
                       tiff => "image/tif",
                       png  => "image/png"
                     );

  ## Create a tnef object
  my %TnefOpts=('output_to_core' => '4194304', 'output_dir' => '/tmp');
  my $tnef = Convert::TNEF->read_ent($ent, \%TnefOpts);
  my $head=$new->head;	# Get the header object from the new message
  if (! $tnef) {
    warn "TNEF CONVERT DID NOT WORK: " . $Convert::TNEF::errstr . "\n";
    warn "  - Failed on msg w/subj: " . $head->get('Subject') . "\n";
    return '';
  }

  #############################################################################
  # This section of code smokes lots of crack, and tries to dig the From:
  # header out of the $tnef->message if the new message we are appending
  # this attachment to does not already have a "From" header.  This is
  # required on most of the Outlook emails that never touch SMTP, only
  # Exchange servers, and never had valid SMTP From headers placed!
  #############################################################################
  my $msg=$tnef->message;
  my $mapi_props=$msg->data('MAPIProps');
  #warn join(", ", keys %{$msg->{MAPIProps}}) . "\n";
  #warn $msg->{MAPIProps}->{MBS_Data} . "\n\n----------------------------\n\n";
  #warn "$mapi_props\n\n---------------------------------\n\n";
  my $test=0x0024;
  #if ($mapi_props =~ m/(\0\0\0\xf8.{20})/) { warn "MATCHED a prop $1\n"; }
  #if (0) {
  if (! length($head->get('From')) ) {
    my $from='';
    my $cntrl_chars='[\c' . join('\c', ('A' .. 'Z')) . ']';
    if ($mapi_props =~ m/From:\s+([^\s\@]+\@[^\s]+)/) {
      $from=$1;
    } elsif ($mapi_props =~ m/\xf8\?\cA\0\0\0$cntrl_chars\0\0\0([^\0]+)\0+\cB\cA/) {
      $from=$1;
    } else {
      if ($mapi_props =~ m/(\xf8\?\cA.{30})/) { warn "MATCH: $1\n"; }
      #$from="Unknown Sender";
    }
    if( length($from)) { $head->replace('from', $from); }
  }
  #############################################################################

  for ($tnef->attachments) {
    $_->longname=~/^[\w\W]+\.([A-Za-z]{2,4})$/;
    my $ext = $1;
    if (!$content_type{lc($ext)}) {
      warn "No MIME type for (" . $_->longname . "/" . $_->name . ")\n";
    } 
    $new->attach( 
                   Type => $content_type{lc($ext)},
                   Encoding => 'base64',
                   Data => $_->data, 
                   Disposition => 'attachment',
                   Filename => $_->longname
                 );
  }
  # If you want to delete the working files
  $tnef->purge;
}

__END__

=pod

# POD documentation

=head1 SYNOPSIS

killtnef --in=source [--out=source]  [--man] [--help] [--version]

=head1 README

killtnef - Converts emails with MS-TNEF, Microsoft's proprietary Transport Neutral Encapsulated Format, attachments into standard MIME-compliant emails.

This script reads an mbox, or a single email message, from STDIN,
extracts data from any ms-tnef attachments that may exist, and writes
a new mbox (or a single email message) to STDOUT which has each of the
files that were encoded in any ms-tnef attachments attached separately,
as RFC-822/RFC-1521 compliant MIME emails/attachments.

Any email(s) containing no ms-tnef MIME attachments are passed through
this script untouched.

When used with NST as an external filter the commandline should look like:

killtnef.exe --in=%IN% --out=%OUT%

=head1 DESCRIPTION

This script was originally written to convert about 35,000 emails from
some Microsoft Outlook *.pst (post office) files, almost all of which
had ms-tnef encoded attachments, into MIME-compliant emails so that
they could be imported into an email-archiving system that 10East
supplies to some of its customers.  If anyone is curious, an imapd
was used to move the emails from the *.pst files to mbox format using
Outlook 2000 as an IMAP client.

This script can also be used as an incoming mail filter which will
automatically convert ms-tnef attachments into MIME-compliant
attachments.

=head1 AUTHORSHIP

Andrew Orr <aorr@10east.com>

Lester Hightower <hightowe@10east.com>

Dennis Heidner (.001%  of the work -- the guys above did all the hard stuff)
=head1 LICENSE

This software is licensed under the terms of the GNU Public License,
which is available for review at http://www.gnu.org/copyleft/gpl.html

=head1 CHANGE LOG

Feb-22-2002: Originally created by Andy Orr

Feb-26-2002: A few enhancements and bug-fixes by Lester Hightower

Mar-06-2002: Documentation and a few comments added by Lester Hightower
in preparation for submitting this script to CPAN.

Mar-07-2002: After realizing that a POD README section is needed for the
HTML pages generated for the script index in CPAN, LHH added one and
submitted this as killtnef-1.0.1.pl.

March-07-2005: Wow, its been three years since anyone had to touch the code.  This version uses
Getopt::Long so we may specify the input file and output file for the conversion.  I've also added
some additional mime types.  Dennis Heidner

April-06-2005: Fixed print to output in all cases, even when no MS TNEF., Dennis Heidner

=head1 PREREQUISITES

This script requires the C<strict> module.
It also requires C<Convert::TNEF 0.16> and C<MIME::Parser 5.406>

=cut